# -*- coding:utf-8 -*-
import logging
logger = logging.getLogger(__name__)
from collections import namedtuple


p = Pair = namedtuple("Pair", "left, right")


class S(object):
    atom = "atom"
    array = "array"
    object = "object"


def model_of(ob):
    from django.db.models.base import ModelBase
    if isinstance(ob, ModelBase):
        return ob
    else:
        return ob.__class__


class Control(object):
    def __init__(self):
        self.fieldmap = {}  # class -> fields

    def get_fieldmap_from_model(self, model):
        try:
            return self.fieldmap[model]
        except KeyError:
            fieldmap = self.fieldmap[model] = {f.name: f for f in model._meta.local_fields}
            for f in model._meta.local_many_to_many:
                fieldmap[f.name] = f
            return fieldmap

    def get_property_from_object(self, ob, k):
        fieldmap = self.get_fieldmap_from_model(model_of(ob))
        try:
            return fieldmap[k]
        except KeyError:
            from django.db.models import IntegerField  # xxx
            fieldmap[k] = IntegerField
            return IntegerField

    def get_type_from_property(self, prop):
        if isinstance(prop, type):
            return prop
        else:
            return prop.__class__

    def get_shape_from_object(self, ob, k):
        if k.endswith("_set"):
            return S.array
        else:
            fieldmap = self.get_fieldmap_from_model(model_of(ob))
            from django.db.models import ManyToManyField
            if isinstance(fieldmap[k], ManyToManyField):
                return S.array
            else:
                return S.object


class Abbreviation(object):
    def __init__(self, control):
        self.control = control

    def __call__(self, ob, name):
        from django.db.models import ForeignKey
        if name == "*":
            for f in ob._meta.local_fields:
                if not isinstance(f, ForeignKey):
                    yield f.name
        else:
            yield name
Empty = ()


class SerializerFactory(object):
    def __init__(self, convertions=None, control=Control(), factory=dict):
        self.convertions = convertions or {}
        self.control = control
        self.factory = factory

    def __call__(self, renaming_options=None, abbreviation=Abbreviation):
        return Serializer(
            self.convertions,
            self.control,
            self.factory,
            renaming_options=renaming_options or {},
            abbreviation=abbreviation(self.control)
        )


class Serializer(object):
    def __init__(self, convertions, control, factory, renaming_options, abbreviation):
        self.convertions = convertions
        self.control = control
        self.factory = factory

        self.abbreviation = abbreviation
        self.renaming_options = renaming_options

    def serialize(self, ob, q_collection, renaming_options=None):
        renaming_options = renaming_options or {}
        r = self.factory()
        for q in q_collection:
            for q in self.abbreviation(ob, q):
                self.build(r, *self.parse(ob, q))
        return r

    def parse(self, ob, q):
        if isinstance(q, Pair):
            k = q.left
            prop = self.control.get_property_from_object(ob, k)
            shape = self.control.get_shape_from_object(ob, k)
            if shape == S.array:
                sub_r = [self.serialize(sub, q.right) for sub in getattr(ob, k).all()]
                return (shape, k, prop, sub_r)
            elif shape == S.object:
                sub_r = self.serialize(getattr(ob, k), q.right)
                return (shape, k, prop, sub_r)
            else:
                raise NotImplemented(shape)
        else:
            return (S.atom, q, self.control.get_property_from_object(ob, q), getattr(ob, q))

    def add_result(self, r, k, v):
        r[self.renaming_options.get(k, k)] = v

    def build(self, r, shape, q, prop, val):
        if shape == S.atom:
            type_ = self.control.get_type_from_property(prop)
            convert = self.convertions.get(type_)
            if convert:
                self.add_result(r, q, convert(val, r))
            else:
                self.add_result(r, q, val)
        elif shape == S.array:
            self.add_result(r, q, val)
        elif shape == S.object:
            self.add_result(r, q, val)
        else:
            raise NotImplemented(shape)



# -*- coding:utf-8 -*-
SECRET_KEY = '@^jwb0=!d03l8x!4hba!oj=#87-f57-oyq%mup^n#x#az_llmq'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ":memory:",
    }
}
INSTALLED_APPS = ["djangosqlash"]



# -*- coding:utf-8 -*-
try:
    import django
except ImportError:
    LIB_ENABLED_DJANGO = False
else:
    LIB_ENABLED_DJANGO = True


def lower_class_name(obj):
    return obj.__class__.__name__.lower()


def get_django_version():
    return django.VERSION


def initialize():
    from django.conf import settings
    settings_dict = dict(
        DATABASES={
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:',
            },
        },
    )
    settings.configure(**settings_dict)


def get_connection():
    "get default database connection"
    from django.db import connections
    connection = connections['default']
    return connection


def get_cursor(connection):
    "get database cursor from connection"
    return connection.cursor()


def get_style():
    from django.core.management.color import no_style
    return no_style()


def create_table(model):
    connection = get_connection()
    cursor = get_cursor(connection)
    style = get_style()

    sql, references = connection.creation.sql_create_model(
        model, style)
    for statement in sql:
        cursor.execute(statement)

    for f in model._meta.many_to_many:
        create_table(f.rel.through)


def _setup_db():
    return get_connection()


def _teardown_db():
    get_connection().close()



# -*- coding:utf-8 -*-

import os
import sys

from setuptools import setup, find_packages
here = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(here, 'README.rst')) as f:
        README = f.read()
    with open(os.path.join(here, 'CHANGES.txt')) as f:
        CHANGES = f.read()
except IOError:
    README = CHANGES = ''

install_requires=[
    'setuptools',
    "django"
    ]

docs_extras = [
    ]

tests_require = [
    ]

testing_extras = tests_require + [
    ]

setup(name='djangosqlash',
      version='0.1',
      description='create dict from object',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        ],
      keywords='',
      author="",
      author_email="",
      url="",
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires = install_requires,
      extras_require = {
          'testing':testing_extras,
          'docs':docs_extras,
          },
      tests_require = tests_require,
      test_suite="djangosqlash.tests",
      entry_points = """      """
      )






# -*- coding:utf-8 -*-
import unittest
from datetime import datetime
import os
os.environ["DJANGO_SETTINGS_MODULE"] = "djangosqlash.tests.settings"
from django.db import models


class Group(models.Model):
    name = models.CharField("NAME", max_length=255)
    created_at = models.DateTimeField(default=lambda: datetime(1999, 12, 31))


class User(models.Model):
    group = models.ForeignKey(Group)
    name = models.CharField(max_length=255)
    created_at = models.DateTimeField(default=lambda: datetime(1999, 12, 31))


class A0(models.Model):
    name = models.CharField(max_length=255)
    created_at = models.DateTimeField(default=lambda: datetime(1999, 12, 31))


class A1(models.Model):
    a0 = models.ForeignKey(A0)
    name = models.CharField(max_length=255)
    created_at = models.DateTimeField(default=lambda: datetime(1999, 12, 31))


class A2(models.Model):
    a1 = models.ForeignKey(A1)
    name = models.CharField(max_length=255)
    created_at = models.DateTimeField(default=lambda: datetime(1999, 12, 31))


class Team(models.Model):
    name = models.CharField("NAME", max_length=255)
    created_at = models.DateTimeField(default=lambda: datetime(1999, 12, 31))


class Member(models.Model):
    teams = models.ManyToManyField(Team)
    name = models.CharField(max_length=255)
    created_at = models.DateTimeField(default=lambda: datetime(1999, 12, 31))


def datetime_for_human(dt, r):
    return dt.strftime("%Y/%m/%d %H:%M:%S")


def int_for_human(v, r):
    return "this is {}".format(v)


class BasicCaseTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from djangosqlash.testing import create_table
        create_table(Group)
        create_table(User)

    def _getTarget(self):
        from djangosqlash import SerializerFactory
        return SerializerFactory

    def _makeOne(self, *args, **kwargs):
        return self._getTarget()(*args, **kwargs)

    def test_simple(self):
        target = self._makeOne({})()
        group = Group(id=1, name="foo")
        result = target.serialize(group, ["name"])
        self.assertEqual(result, {"name": "foo"})

    def test_convert(self):
        from datetime import datetime
        from django.db.models import DateTimeField
        target = self._makeOne({DateTimeField: datetime_for_human})()
        group = Group(id=1, name="foo", created_at=datetime(2000, 1, 1))
        result = target.serialize(group, ["created_at"])
        self.assertEqual(result, {'created_at': '2000/01/01 00:00:00'})

    def test_foreignkey(self):
        from django.db.models import IntegerField
        user = User(id=1)
        user.group_id = 1
        target = self._makeOne({IntegerField: int_for_human})()
        result = target.serialize(user, ["group_id"])
        self.assertEqual(result, {'group_id': 'this is 1'})

    def test_relation_onetomany(self):
        from datetime import datetime
        from djangosqlash import Pair

        target = self._makeOne()()
        group = Group(id=1, name="foo")
        group.user_set.add(User(id=1, group=group, name="boo", created_at=datetime(2000, 1, 1)))
        group.user_set.add(User(id=2, group=group, name="yoo", created_at=datetime(2000, 1, 1)))
        result = target.serialize(group, ["name", Pair("user_set", ["name"])])
        self.assertEqual(result, {'user_set': [{'name': 'boo'}, {'name': 'yoo'}], 'name': 'foo'})

    def test_relation_manytoone(self):
        from datetime import datetime
        from djangosqlash import Pair

        target = self._makeOne()()
        group = Group(id=1, name="foo")
        group.user_set.add(User(id=1, group=group, name="boo", created_at=datetime(2000, 1, 1)))
        group.user_set.add(User(id=2, group=group, name="yoo", created_at=datetime(2000, 1, 1)))
        result = target.serialize(group.user_set.get(pk=1), ["name", Pair("group", ["name"])])
        self.assertEqual(result, {'group': {'name': 'foo'}, 'name': 'boo'})

    def test_abbreviation(self):
        from datetime import datetime
        from django.db.models import IntegerField, DateTimeField, AutoField

        user = User(name="foo", created_at=datetime(2000, 1, 1))
        target = self._makeOne({IntegerField: int_for_human,
                                AutoField: int_for_human,
                                DateTimeField: datetime_for_human})()
        result = target.serialize(user, ["*"])
        self.assertEqual(result, {'name': 'foo', 'created_at': '2000/01/01 00:00:00', 'id': 'this is None'})

    def test_renaming(self):
        from datetime import datetime
        from django.db.models import IntegerField, DateTimeField, AutoField

        user = User(name="foo", created_at=datetime(2000, 1, 1))
        factory = self._makeOne({IntegerField: int_for_human,
                                AutoField: int_for_human,
                                DateTimeField: datetime_for_human})
        target = factory({"name": "Name", "created_at": "CreatedAt", "id": "Id"})
        result = target.serialize(user, ["*"])
        self.assertEqual(result, {'Name': 'foo', 'CreatedAt': '2000/01/01 00:00:00', 'Id': 'this is None'})


class NestedCaseTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from djangosqlash.testing import create_table
        create_table(A0)
        create_table(A1)
        create_table(A2)

    def _getTarget(self):
        from djangosqlash import SerializerFactory
        return SerializerFactory

    def _makeOne(self, *args, **kwargs):
        return self._getTarget()(*args, **kwargs)

    def test_deep_nested(self):
        from django.db.models import DateTimeField
        from datetime import datetime
        from djangosqlash import Pair

        target = self._makeOne({DateTimeField: datetime_for_human})()
        a2 = A2(id=1, a1=A1(id=1, a0=A0(id=1, created_at=datetime(2000, 1, 1))))
        result = target.serialize(a2, [Pair("a1", [Pair("a0", ["created_at"])])])
        assert result == {'a1': {'a0': {'created_at': '2000/01/01 00:00:00'}}}


class ManyToManyTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from djangosqlash.testing import create_table, _setup_db
        _setup_db()
        create_table(Team)
        create_table(Member)

    def _getTarget(self):
        from djangosqlash import SerializerFactory
        return SerializerFactory

    def _makeOne(self, *args, **kwargs):
        return self._getTarget()(*args, **kwargs)

    def test_many_to_many(self):
        from django.db.models import DateTimeField
        from djangosqlash import Pair

        target = self._makeOne({DateTimeField: datetime_for_human})()

        team0 = Team(id=1, name="foo"); team0.save()
        team1 = Team(id=2, name="boo"); team1.save()
        member0 = Member(id=1, name="x"); member0.save()
        member1 = Member(id=2, name="y"); member1.save()
        member2 = Member(id=3, name="z"); member2.save()
        team0.member_set.add(member0)
        team0.member_set.add(member1)
        team1.member_set.add(member1)
        team1.member_set.add(member2)
        result = target.serialize(team0, ["name", "created_at", Pair("member_set", ["name", "created_at"])])
        self.assertEqual(result, {'created_at': '1999/12/31 00:00:00',
                                  'member_set': [{'created_at': '1999/12/31 00:00:00', 'name': 'x'},
                                                 {'created_at': '1999/12/31 00:00:00', 'name': 'y'}], 'name': 'foo'})

    def test_many_to_many_reverse(self):
        from django.db.models import DateTimeField
        from djangosqlash import Pair

        target = self._makeOne({DateTimeField: datetime_for_human})()

        team0 = Team(id=1, name="foo"); team0.save()
        team1 = Team(id=2, name="boo"); team1.save()
        member0 = Member(id=1, name="x"); member0.save()
        member1 = Member(id=2, name="y"); member1.save()
        member2 = Member(id=3, name="z"); member2.save()
        team0.member_set.add(member0)
        team0.member_set.add(member1)
        team1.member_set.add(member1)
        team1.member_set.add(member2)
        result = target.serialize(member1, ["name", "created_at", Pair("teams", ["name", "created_at"])])
        self.assertEqual(result, {'name': 'y',
                                  'created_at': '1999/12/31 00:00:00',
                                  'teams': [{'name': 'foo', 'created_at': '1999/12/31 00:00:00'},
                                            {'name': 'boo', 'created_at': '1999/12/31 00:00:00'}]})
